/*
* Copyright(c) 1997-2001 Id Software, Inc.
* Copyright(c) 2002 The Quakeforge Project.
* Copyright(c) 2006 Quetoo.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or(at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
* 
* See the GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "client.h"

typedef struct loc_s {
	vec3_t loc;
	char desc[MAX_STRING_CHARS];
} loc_t;

#define MAX_LOCATIONS 1024

loc_t locations[MAX_LOCATIONS];
int numlocations;

/*
LOC_ClearLocations

Effectively clears all locations for the current level.
*/
void LOC_ClearLocations(void){
	numlocations = 0;
}

/*
LOC_LoadLocations

Parse a .loc file for the current level.
*/
void LOC_LoadLocations(void){
	char locfile[MAX_QPATH];
	FILE *f;
	int i, err;
	
	LOC_ClearLocations();  // clear any resident locations
	err = i = 0;
	
	strcpy(locfile, cl.configstrings[CS_MODELS + 1]); // start with .bsp name
	strcpy(locfile + strlen(locfile) - 3, "loc"); // and change to .loc
	
	if(FS_FOpenFile(locfile, &f) == -1){
		Com_Printf("Couldn't load %s\n", locfile);
		return; // no loc file exists for this level
	}
	
	while(i < MAX_LOCATIONS && err != EOF){
		
		err = fscanf(f, "%f %f %f %[a-zA-Z ]",
				&locations[i].loc[0], &locations[i].loc[1],
				&locations[i].loc[2], locations[i].desc
		);
		
		numlocations = i;
		i++;
	}
	
	
	Com_Printf("Loaded %i locations.\n", numlocations);
	FS_FCloseFile(f);
}

/*
LOC_SaveLocations_f

Write locations for current level to file.
*/
void LOC_SaveLocations(void){
	char locfile[MAX_QPATH];
	FILE *f;
	int i;
	
	snprintf(locfile, MAX_QPATH, "%s/%s", FS_Gamedir(), cl.configstrings[CS_MODELS + 1]);
	strcpy(locfile + strlen(locfile) - 3, "loc"); // change to .loc
	
	if((f = fopen(locfile, "w")) == NULL){
		Com_Printf("Could not write %s\n", locfile);
		return;
	}
	
	for(i = 0; i < numlocations; i++){
		fprintf(f, "%d %d %d %s\n",
				(int)locations[i].loc[0], (int)locations[i].loc[1],
				(int)locations[i].loc[2], locations[i].desc
		);
	}
	
	Com_Printf("Saved %d locations.\n", numlocations);
	FS_FCloseFile(f);
}

/*
LOC_Location

Returns the description of the location nearest near.
*/
char *LOC_Location(vec3_t near){
	vec_t dist, mindist = 9999999;
	vec3_t v;
	int i, j;
	
	if(numlocations == 0)
		return "";
	
	for(i = 0, j = 0; i < numlocations; i++){ // find closest loc
		
		VectorSubtract(near, locations[i].loc, v);
		if((dist = VectorNormalize(v)) < mindist){ // closest yet
			mindist = dist;
			j = i;
		}
	}
	
	return locations[j].desc;
}

/*
LOC_LocationHere

Returns the description of the location nearest the client.
*/
char *LOC_LocationHere(void){
	return LOC_Location(cl.viewdef.vieworg);
}

// we borrow tracing from cmodel to do $loc_there
extern trace_t CM_BoxTrace(vec3_t start, vec3_t end, 
		vec3_t mins, vec3_t maxs, int headnode, int brushmask);

/*
LOC_LocationThere

Returns the description of the location nearest the client's crosshair.
*/
char *LOC_LocationThere(void){
	vec3_t forward, end;
	trace_t tr;
	
	// project vector from view position and angle
	AngleVectors(cl.viewdef.viewangles, forward, NULL, NULL);
	VectorMA(cl.viewdef.vieworg, 8192, forward, end);
	
	// and trace to world model
	tr = CM_BoxTrace(cl.viewdef.vieworg, end,
			vec3_origin, vec3_origin, 0, MASK_SHOT
	);
	
	return LOC_Location(tr.endpos);
}

/*
LOC_AddLocation

Add a new location described by desc at near.
*/
void LOC_AddLocation(vec3_t near, char *desc){
	
	if(numlocations >= MAX_LOCATIONS)
		return;
	
	VectorCopy(near, locations[numlocations].loc);
	strncpy(locations[numlocations].desc, desc, MAX_STRING_CHARS);
	
	numlocations++;
}

/*
LOC_AddLocation_f

Command callback for adding locations in game.
*/
void LOC_AddLocation_f(void){
	
	if(Cmd_Argc() < 2){
		Com_Printf("Usage: addloc <description>\n");
		return;
	}
	
	LOC_AddLocation(cl.viewdef.vieworg, Cmd_Args());
}

/*
LOC_Init
*/
void LOC_Init(void){
	Cmd_AddCommand("addloc", LOC_AddLocation_f);
	Cmd_AddCommand("savelocs", LOC_SaveLocations);
}
